#ifndef BRANCHOBJ_H
#define BRANCHOBJ_H

#include "floatimageobj.h"
#include "linkablemapobj.h"
#include "ornamentedobj.h"
#include "xlinkobj.h"


bool isAbove(BranchObj*,BranchObj*);

/*! \brief A branch visible in the map */

/*! If HideExport is used, this branch and its children will be hidden in export */
enum HideTmpMode {HideNone, HideExport};


/////////////////////////////////////////////////////////////////////////////
class BranchObj:public OrnamentedObj {
public:
	/*! New branches will get use same color for heading as parent */
	enum BranchModification {NewBranch, MovedBranch};


    BranchObj ();
    BranchObj (QGraphicsScene*);
    BranchObj (QGraphicsScene*, LinkableMapObj* parent);
    ~BranchObj ();
	bool operator< ( const BranchObj & );
	bool operator== ( const BranchObj & );
    virtual void init ();
    virtual void copy (BranchObj*);
    void clear();

    virtual void setParObjTmp (LinkableMapObj*,QPointF,int);// Only for moving Obj around
	virtual void unsetParObjTmp();			// reuse original ParObj

	virtual void setVisibility(bool,int);	// set visibility
    virtual void setVisibility(bool);	    // set vis. for w
	virtual void setLinkColor();			// set the color of link
	virtual void setColorSubtree(QColor);	// set the color of heading

	virtual void positionContents();
    virtual void move (double x,double y);
    virtual void move (QPointF);
    virtual void moveBy (double x,double y);
    virtual void moveBy (QPointF);
    virtual void positionBBox();
    virtual void calcBBoxSize();
	virtual void setDockPos();
    virtual LinkableMapObj* findMapObj(QPointF,LinkableMapObj*);	// find MapObj 
    virtual LinkableMapObj* findID (QString sid);	// find Obj by ID string
    virtual void setHeading (QString);

	virtual void setHideTmp (HideTmpMode);
	virtual bool hasHiddenExportParent ();

	virtual QString saveToDir (const QString&,const QString&, const QPointF&);// Save data recursivly to tempdir

	virtual void addXLink (XLinkObj*);
	virtual void removeXLinkRef (XLinkObj*);// Remove ref in list
	virtual void deleteXLink (XLinkObj*);	// remove references and delete XLinkObj 
	virtual void deleteXLinkAt (int);		// remove references and delete XLinkObj 
	virtual XLinkObj* XLinkAt (int);		// return reference of XLinkObj 
	virtual BranchObj* XLinkTargetAt (int);

	void setIncludeImagesVer(bool);
	bool getIncludeImagesVer();
	void setIncludeImagesHor(bool);
	bool getIncludeImagesHor();
	QString getIncludeImageAttr();

	virtual FloatImageObj* addFloatImage();
	virtual FloatImageObj* addFloatImage(FloatImageObj*);
	virtual void removeFloatImage(FloatImageObj*);
    virtual FloatImageObj* getFirstFloatImage();
    virtual FloatImageObj* getLastFloatImage();
	virtual FloatImageObj* getFloatImageNum(const uint &);
protected:	
	virtual void savePosInAngle();					// write pos in angle for resorting			
	virtual void setDefAttr (BranchModification);	// set default attributes (font, size, ...)
public:	
    virtual BranchObj* addBranch();
    virtual BranchObj* addBranch(BranchObj*);		// makes deep copy of BranchObj
    virtual BranchObj* addBranchPtr(BranchObj*);	// just adds pointer
    virtual BranchObj* insertBranch(int);
    virtual BranchObj* insertBranch(BranchObj*,int);
    virtual BranchObj* insertBranchPtr (BranchObj*,int);
    virtual void removeBranchHere(BranchObj*);  
    virtual void removeChildren();  
    virtual void removeBranch(BranchObj*);  
    virtual void removeBranchPtr (BranchObj*);  

    virtual bool canMoveBranchUp();
    virtual BranchObj* moveBranchUp(BranchObj*);
    virtual bool canMoveBranchDown();
    virtual BranchObj* moveBranchDown(BranchObj*);

    virtual void sortChildren();
    virtual BranchObj* linkTo (BranchObj*, int);
    virtual void alignRelativeTo(const QPointF, bool alignSelf=false );
	virtual void reposition();
	virtual void unsetAllRepositionRequests();

	virtual QPolygonF shape();				//!< Returns arbitrary bounding polygon
	virtual QRectF getTotalBBox();			// return BBox including children			
	virtual QRectF getBBoxSizeWithChildren();	// return size of BBox including children  
	virtual void calcBBoxSizeWithChildren();	// calc size of  BBox including children recursivly

	virtual QString getSelectString();
	virtual void setAnimation(const AnimPoint &ap);
	virtual bool animate();

protected:
	static BranchObj* itLast;		// iterator for first(), next()
	static BranchObj* itFirst;		// first iterator for first(), next()
    QList<BranchObj*> branch;		// all child branches
	QList<FloatImageObj*> floatimage;// child images
	QList<XLinkObj*> xlink;			// xlinks to other branches

	AnimPoint anim;

public:	
	float angle;					// used in mainbranch to reorder mainbranches
protected:	
//	bool scrolled;					// true if all children are scrolled and thus invisible
//	bool tmpUnscrolled;				// can only be true (temporary) for a scrolled subtree
	bool includeImagesVer;			// include floatimages in bbox vertically
	bool includeImagesHor;			// include floatimages in bbox horizontally
};


#endif

