#include <iostream>
using namespace std;

#include <QStringList>

#include "treeitem.h"

#include "branchobj.h"
#include "branchitem.h"
#include "vymmodel.h"

TreeItem::TreeItem(const QList<QVariant> &data, TreeItem *parent)
{
	model=NULL;

    parentItem = parent;
    itemData = data;
	lmo=NULL;

	branchOffset=0;
	branchCounter=0;
	lastSelectedBranchNum=-1;

	note.setNote(""); 
	// note.setFontHint (textEditor->getFontHintDefault() );	//FIXME-2
	// isNoteInEditor=false;
}

TreeItem::~TreeItem()
{
    qDeleteAll(childItems);
}

QString TreeItem::saveToDir (const QString &tmpdir,const QString &prefix, const QPointF& offset)
{
	qWarning ("TreeItem::saveToDir called directly");
	return QString();
}


/*
void TreeItem::copy (OrnamentedObj* other)	//FIXME-2

{
	note.copy (other->note);
	model=other->model;
}
*/

void TreeItem::setModel (VymModel *m)
{
	model=m;
}

VymModel* TreeItem::getModel ()
{
	return model;
}

void TreeItem::appendChild(TreeItem *item)
{
    childItems.append(item);
	item->setModel (model);

	if (item->type == Branch || item->type ==MapCenter)
	{
		if (branchCounter==0)
			branchOffset=childItems.count()-1;
		branchCounter++;
	}
}

void TreeItem::removeChild(int row)
{
    if (row<0 || row > childItems.size()-1)
		qWarning ("TreeItem::removeChild tried to remove non existing item?!\n");
	else
	{
		cout << "TI::removeChild this="<<this<<"  row="<<row<<endl;
		delete childItems.takeAt (row);
	}
}

TreeItem *TreeItem::child(int row)
{
    return childItems.value(row);
}

int TreeItem::childCount() const
{
    return childItems.count();
}

int TreeItem::childNumber() const
{
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<TreeItem*>(this));

    return 0;
}

int TreeItem::columnCount() const
{
    return itemData.count();
}

int TreeItem::branchCount() const
{
    return branchCounter;
}

int TreeItem::imageCount() const
{
	int imageCounter=0;
    return imageCounter; // FIXME-1 imageCounter needs to be calculated...
}

int TreeItem::xlinkCount() const // FIXME-2 check if xlinks are stored in a different way (global to model?)
{
	int xlinkCounter=0;
    return xlinkCounter; // FIXME-1 xlinkCounter needs to be calculated...
}

int TreeItem::row() const
{
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<TreeItem*>(this));

    return 0;
}

int TreeItem::column() const
{
    return 0;
}

int TreeItem::depth() 
{
	int d=-1;
	TreeItem *ti=this;
	while (ti!=NULL)
	{
		ti=ti->parent();
		d++;
	}
	return d;
}

TreeItem *TreeItem::parent()
{
    return parentItem;
}

int TreeItem::num()
{
	switch (type)
	{
		case Undefined: return -1;
		case MapCenter: return parentItem->childItems.indexOf (this) - branchOffset;
		case Branch: return parentItem->childItems.indexOf (this) - branchOffset;
		case Image: return -1; // FIXME-2
		default: return -1;
	}
}

int TreeItem::num (TreeItem *item)
{
	if (!item) return -1;
	switch (item->getType())
	{
		case Undefined: return -1;
		case MapCenter: return childItems.indexOf (this) - branchOffset;
		case Branch: return childItems.indexOf (this) - branchOffset;
		case Image: return -1; // FIXME-2
		default: return -1;
	}
}

QVariant TreeItem::data(int column) const
{
    return itemData.value(column);
}

void TreeItem::setHeading (const QString s)
{
	itemData[0]=s;
}

QString TreeItem::getHeading ()
{
	return itemData[0].toString();
}

void TreeItem::setNote(const QString s)
{
	note.setNote(s);
	// updateNoteFlag();  FIXME-2 needed?
	if (model) 
		model->emitContentHasChanged (this);
}

void TreeItem::setNoteObj(const NoteObj &n, bool updateNoteEditor)
{
	note=n;
	// updateNoteFlag();  FIXME-2 needed?
	if (model && updateNoteEditor)	
		model->emitContentHasChanged (this);
}

QString TreeItem::getNote()
{
	return note.getNote();
}

NoteObj TreeItem::getNoteObj()
{
	return note;
}

QString TreeItem::getNoteASCII(const QString &indent, const int &width)
{
    return note.getNoteASCII(indent,width);
}

QString TreeItem::getNoteASCII()
{
    return note.getNoteASCII();
}

QString TreeItem::getNoteOpenDoc()
{
    return note.getNoteOpenDoc();
}

void TreeItem::setType(const Type t)
{
	type=t;
	itemData[1]=getTypeName();
}

TreeItem::Type TreeItem::getType()
{
	return type;
}

QString TreeItem::getTypeName()
{
	switch (type)
	{
		case Undefined: return QString ("Undefined");
		case MapCenter: return QString ("MapCenter");
		case Branch: return QString ("Branch");
		case Image: return QString ("Image");
		default: return QString ("TreeItem::getTypeName no typename defined?!");
	}
}

TreeItem* TreeItem::getChildNum(const int &n)
{
	if (n>=0 && n<childItems.count() )
		return childItems.at(n);
	else
		return NULL;
}

BranchItem* TreeItem::getFirstBranch()
{
	if (branchCounter>0)
		return getBranchNum (branchOffset);
	else
		return NULL;
}

BranchItem* TreeItem::getLastBranch()
{
	if (branchCounter>0)
		return getBranchNum (branchOffset + branchCounter-1);
	else
		return NULL;
}


BranchItem* TreeItem::getBranchNum(const int &n)
{
	if (branchCounter>0)
		return (BranchItem*)getChildNum (branchOffset + n);
	else
		return NULL;
}

BranchObj* TreeItem::getBranchObjNum(const int &n)
{
	if (branchCounter>0)
	{
		return (BranchObj*)(getChildNum (branchOffset+n)->lmo);
	} else
		return NULL;
}

void TreeItem::setLastSelectedBranch()
{
	if (parentItem)
		parentItem->lastSelectedBranchNum=parentItem->childItems.indexOf(this);
}

TreeItem* TreeItem::getLastSelectedBranch()
{
	return getBranchNum (lastSelectedBranchNum);
}

LinkableMapObj* TreeItem::getLMO()	// FIXME-3  VM should be unnecessary in the end
{
	return lmo;
}

void TreeItem::setLMO(LinkableMapObj *l)
{
	lmo=l;
}

