#ifndef TREEITEM_H
#define TREEITEM_H

#include <QList>
#include <QVariant>

#include "noteobj.h"
#include "xmlobj.h"

class LinkableMapObj;
class BranchObj;
class BranchItem;
class VymModel;

class TreeItem:public XMLObj
{
public:
	enum Type {Undefined,MapCenter,Branch,Image};
    TreeItem(const QList<QVariant> &data, TreeItem *parent = 0);
    ~TreeItem();
	QString saveToDir (const QString &tmpdir,const QString &prefix, const QPointF& offset);


	// General housekeeping
	void setModel (VymModel *m);
	VymModel* getModel();

    void appendChild (TreeItem *child);
	void removeChild (int row);

    TreeItem *child(int row);
    int childCount() const;
    int childNumber() const;
    int columnCount() const;
	int branchCount() const;
	int imageCount() const;
	int xlinkCount() const;

    int row() const;
	int column() const;
	int depth() ;
    TreeItem *parent();

	/*! Return number of item in parent by type, 
	    e.g. first branch has number 0           */
	int num();		
	int num (TreeItem *item);	//! Return number of item by type

	// Accessing data
    QVariant data(int column) const;
	void setHeading (const QString s);
	QString getHeading();

protected:
	NoteObj note;
public:	
	void setNote(const QString s);
	QString getNote();
	virtual void setNoteObj(const NoteObj &, bool updateNoteEditor=true);
	virtual NoteObj getNoteObj();			
	virtual QString getNoteASCII(const QString &indent, const int &width); // returns note	(ASCII)
    virtual QString getNoteASCII();			// returns note	(ASCII)
    virtual QString getNoteOpenDoc();		// returns note	(OpenDoc)


protected:
	Type type;
public:	
	void setType (const Type t);
	Type getType ();
	QString getTypeName ();

	// Navigation and selection
	TreeItem* getChildNum(const int &n);
	BranchItem* getFirstBranch();
	BranchItem* getLastBranch();
	BranchItem* getBranchNum(const int &n);
	BranchObj* getBranchObjNum(const int &n);
	void setLastSelectedBranch();
	TreeItem* getLastSelectedBranch();

	
	// Relation to map objects in graphicsscene
	LinkableMapObj* getLMO();
	void setLMO (LinkableMapObj*);

protected:
	VymModel *model;

    QList<TreeItem*> childItems;
    QList<QVariant> itemData;
    TreeItem *parentItem;
 
	LinkableMapObj *lmo;

	int branchOffset;
	int branchCounter;
	int lastSelectedBranchNum;

	bool hidden;	//! Hidden in export if true
};

#endif
