#include <QtGui>

#include <iostream>
using namespace std;

#include "branchitem.h"
#include "treeitem.h"
#include "treemodel.h"

TreeModel::TreeModel(QObject *parent)
    : QAbstractItemModel(parent)
{
    QList<QVariant> rootData;
    rootData << "Heading" << "Type" <<"Note";
    rootItem = new TreeItem(rootData);
}

TreeModel::~TreeModel()
{
    delete rootItem;
}

QVariant TreeModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return QVariant();

    if (role != Qt::DisplayRole)
        return QVariant();

    TreeItem *item = getItem (index);

    return item->data(index.column());
}

Qt::ItemFlags TreeModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return Qt::ItemIsEnabled;

    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant TreeModel::headerData(int section, Qt::Orientation orientation,
                               int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return rootItem->data(section);

    return QVariant();
}

QModelIndex TreeModel::index(int row, int column, const QModelIndex &parent)
            const
{
    TreeItem *parentItem;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = getItem (parent);

    TreeItem *childItem = parentItem->child(row);
    if (childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

QModelIndex TreeModel::parent(const QModelIndex &index) const
{
    if (!index.isValid())
        return QModelIndex();

    TreeItem *ti= getItem (index);
    TreeItem *parentItem = ti->parent();

	//cout << "TreeModel::parent  ti="<<ti<<" "<<ti->getHeading().toStdString()<<"  pi="<<parentItem<<"  "<<endl;
    if (parentItem == rootItem)
        return QModelIndex();

	if (!parentItem)
	{
		cout <<"TreeModel::parent      ti=="<<ti<<"  "<<ti->getHeading().toStdString()<<endl;
        return QModelIndex();	// FIXME-3 do this to avoid segfault, but why?
		                        // see also my question on qt-interest in march
	}	

    return createIndex(parentItem->childNumber(), 0, parentItem);
}

int TreeModel::rowCount(const QModelIndex &parent) const
{
    TreeItem *parentItem;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = getItem (parent);

    return parentItem->childCount();
}

int TreeModel::columnCount(const QModelIndex &parent) const
{
    if (parent.isValid())
        return getItem (parent)->columnCount();
    else
        return rootItem->columnCount();
}

BranchItem* TreeModel::next(BranchItem* &current, BranchItem* &previous, int &d0)
{
	// Walk through map beginning at current with previous==0
	// Start at root, if current==NULL
	if (!current) current=(BranchItem*)rootItem;

	// Are we just beginning to walk the map?
	if (!previous)
	{
		previous=current;
		d0=current->depth();
		current=current->getFirstBranch();
		return current;
	}

	//std::cout << " cur="<<current->getHeading().toStdString();
	//std::cout << " prev="<<previous->getHeading().toStdString()<<std::endl;

	// Going up or down (deeper)?
	if (current->depth() > previous->depth() )
	{	
		// Coming from above
		// Trying  to go down deeper
		if (current->branchCount() >0 )
		{
			previous=current;
			current=current->getFirstBranch();
			return current;
		}	
		// turn around and go up again
	}	

	// Coming from below,
	// Trying to go down again to siblings

	BranchItem *sibling=current->getBranchNum (previous->num()+1);

	if (sibling)
	{	
		// Found sibling of previous, go there
		previous=current;
		current=sibling;
		return current;
	} 

	// Go up and try to find siblings of current
	previous=current;
	current=(BranchItem*)current->parent();

	// Check if we still can go somewhere
	if (!current) return current;
	
	while (current && current->depth() < previous->depth() )
		next (current,previous,d0);
		
	return current;
}

bool TreeModel::insertRows ( int row, int count, const QModelIndex & parent)
{
	std::cout << "TreeModel insertRows()\n";
	int last=row+count-1;
	beginInsertRows (parent,row,last);

	for (int i=row; i<=last; i++)
	{
		std::cout << "TreeModel::insertRows inserting i="<<i<<std::endl;
	}
	endInsertRows ();
	return true;
}

bool TreeModel::removeRows ( int row, int count, const QModelIndex & parent)
{
	int last=row+count-1;
	cout << "TreeModel::removeRows row="<<row<<"  count="<<count<<"  last="<<last<<endl;
    TreeItem *pi= getItem (parent);
	cout << "  pi="<<pi<<"  "<<pi->getHeading().toStdString()<<endl;
	beginRemoveRows (parent,row,last);

	for (int i=row; i<=last; i++)
	{
		cout << "TreeModel::removeRows removing i="<<i<<std::endl;
		pi->removeChild (row);
	}
	endRemoveRows ();
	return true;
}

TreeItem *TreeModel::getItem(const QModelIndex &index) const
{
    if (index.isValid()) {
		TreeItem *item = static_cast<TreeItem*>(index.internalPointer());

        if (item) return item;
    }
    return rootItem;
}

QModelIndex TreeModel::index (TreeItem* ti)
{
	return createIndex (ti->row(),ti->column(),ti);
}

