#include "vymview.h"

#include <iostream>
using namespace std;

#include "mainwindow.h"
#include "mapeditor.h"

extern Main *mainWindow;


VymView::VymView(VymModel *m)
{
	model=m;

	// Create TreeView
	treeview=new QTreeView;
	treeview->setModel ((QAbstractItemModel*)model);
	treeview->setMinimumWidth (350);
	treeview->setColumnWidth (0,350);

	selModel=treeview->selectionModel();
	model->setSelectionModel (selModel);
	connect (
		selModel, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		this,SLOT (changeSelection(const QItemSelection &,const QItemSelection &)));

	// Create good old MapEditor
	MapEditor* me=model->getMapEditor();
	if (!me) me=new MapEditor (model);
	connect (
		selModel, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		me,SLOT (updateSelection(const QItemSelection &,const QItemSelection &)));
	connect (
		selModel, SIGNAL (currentChanged(const QModelIndex &, const QModelIndex &)), 
		me,SLOT (updateCurrent(const QModelIndex &,const QModelIndex &)));

	// VymModel may want to update selection or other data, e.g. during animation
	connect (
		model, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		me,SLOT (updateSelection(const QItemSelection &,const QItemSelection &)));
	connect (
		model, SIGNAL (newChildObject(QModelIndex) ),
		this,SLOT (updateChilds (QModelIndex) ) );
	connect (
		model, SIGNAL (contentHasChanged(QModelIndex) ),
		mainWindow, SLOT (updateContent(QModelIndex) ) );
		

	//me->viewport()->setFocus();	//FIXME-3 needed?
	me->setAntiAlias (mainWindow->isAliased());
	me->setSmoothPixmap(mainWindow->hasSmoothPixmapTransform());

	addWidget (treeview);
	addWidget (me);

	// Set geometry
	QList <int> sizes;
	sizes.append (120);
	sizes.append (600);
	setSizes (sizes);
}

QItemSelectionModel* VymView::selectionModel() 
{
	if (treeview) 
		return selModel;
	else 
		std::cout <<"VymView::selectionModel: hey, no treeview so far???\n";
	return NULL;
}


void VymView::updateChilds (QModelIndex ix)
{
	treeview->setExpanded (ix,true);
}

void VymView::changeSelection (const QItemSelection &newsel, const QItemSelection &oldsel)
{
	/*
	cout <<"VymView::changeSelection (";
	if (!newsel.indexes().isEmpty() )
		cout << model->getItem(newsel.indexes().first() )->getHeading().toStdString();
	cout << " <- ";
	if (!oldsel.indexes().isEmpty() )
		cout << model->getItem(oldsel.indexes().first() )->getHeading().toStdString();
	cout << ")\n";
	*/

	// Notify mainwindow to update satellites like NoteEditor, if needed (model==currenModel...)
	mainWindow->changeSelection (model,newsel,oldsel);	// FIXME-3 maybe connect VymModel <-> MainWindow directly?
}

